;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/Llib/thread.scm          */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Feb  4 11:49:11 2002                          */
;*    Last change :  Fri May 23 07:10:16 2003 (serrano)                */
;*    Copyright   :  2002-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    Bigloo Fair threads.                                             */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_thread

   (with   __ft_eval)
   
   (import __ft_types
	   __ft_%types
	   __ft_%thread
	   __ft_scheduler
	   __ft_%scheduler
	   __ft_env
	   __ft_%env
	   __ft_signal
	   __ft_reader
	   __ft_writer
	   __ft_socket
	   __ft_process
	   __ft_exception
	   __ft_%exception
	   __ft_%mutex)
   
   (static (class %thread::thread)
	   
	   (class %thread-term-sgn
	      (thread::thread read-only)))
   
   (export (make-thread::thread ::procedure . name)
	   (thread-sleep! ::int)
	   (thread-await! ::obj . timeout)
	   (thread-await*! ::pair . timeout)
	   (thread-await-values! ::obj . timeout)
	   (thread-await-values*! ::pair . timeout)
	   (thread-join! ::thread . timeout)
	   (current-thread)
	   (thread-start! ::thread . scheduler)
	   (thread-terminate! ::thread)
	   (thread-suspend! ::thread)
	   (thread-resume! ::thread)
	   (thread-yield!)
	   (thread-get-values! ::obj)
	   (thread-get-values*! ::pair)))

;*---------------------------------------------------------------------*/
;*    object-equal? ::%thread-term-sgn ...                             */
;*---------------------------------------------------------------------*/
(define-method (object-equal? o1::%thread-term-sgn o2)
   (and (%thread-term-sgn? o2)
	(eq? (%thread-term-sgn-thread o1)
	     (%thread-term-sgn-thread o2))))

;*---------------------------------------------------------------------*/
;*    make-thread ...                                                  */
;*---------------------------------------------------------------------*/
(define (make-thread runnable . name)
   (define (thread-body t)
      (with-access::%thread t (%state %result %mutexes cleanup scheduler)
	 ;; terminate is used to abruptly terminate a thread
	 (let ((res (bind-exit (terminate)
		       (%thread-%terminate-set! t terminate)
		       (try (runnable)
			    (lambda (escape obj proc msg)
			       (notify-error obj proc msg)
			       (terminate #f))))))
	    ;; mark that the thread is now dead
	    (thread-set-dead! t)
	    ;; store the result of the thread
	    (set! %result res)
	    ;; abandon all the mutexes owned by the thread
  	    (for-each abandon-mutex! %mutexes)
	    (set! %mutexes '())
	    ;; raise a signal for the thread termination
	    (broadcast! (instantiate::%thread-term-sgn (thread t)) res)
	    ;; invoke the thread cleanup
	    (if (procedure? cleanup)
		(if (correct-arity? cleanup 1)
		    (cleanup res)
		    (error t "Illegal cleanup function" cleanup)))
	    ;; switch back to the scheduler
	    (thread-switch-then-exit t))))
   (if (not (correct-arity? runnable 0))
       (error "make-thread"
	      "Illegal thread body (should be a procedure of 0 arguments)"
	      runnable)
       (let ((the-name (if (null? name)
			   (gensym 'fairthread)
			   (cond
			      ((symbol? (car name))
			       (car name))
			      ((string? (car name))
			       (string->symbol (car name)))
			      (else
			       (gensym 'fairthread))))))
	  (letrec ((t (instantiate::%thread
			 (%builtin (%bgl-thread-new/name
				    (lambda () (thread-body t))
				    the-name))
			 (name (if (pair? name) (car name) the-name))
			 (%exc-handlers (list uncaught-exception-hdl)))))
	     (thread-setup! t)
	     t))))

;*---------------------------------------------------------------------*/
;*    thread-join! ...                                                 */
;*---------------------------------------------------------------------*/
(define (thread-join! t . args)
   (with-access::thread t (%state %result %exc-result %exc-raised)
      (if (or (eq? %state 'dying) (eq? %state 'dead))
	  (if %exc-raised
	      (raise %exc-result)
	      %result)
	  (match-case args
	     ((?to ?to-val)
	      (if (not (number? to))
		  (bigloo-type-error "thread-join" "integer" to)
		  (let ((v (thread-await! (instantiate::%thread-term-sgn
					     (thread t))
					  to)))
		     (cond
			((terminated-thread-exception? (thread-%exc-result t))
			 (raise (thread-%exc-result t)))
			(v
			 v)
			(else
			 to-val)))))
	     ((?to)
	      (if (not (number? to))
		  (bigloo-type-error "thread-join" "integer" to)
		  (let ((v (thread-await! (instantiate::%thread-term-sgn
					     (thread t))
					  to)))
		     (cond
			((terminated-thread-exception? (thread-%exc-result t))
			 (raise (thread-%exc-result t)))
			(v
			 v)
			(else
			 (raise *join-timeout-exception*))))))
	     (else
	      (let ((v (thread-await! (instantiate::%thread-term-sgn
					 (thread t)))))
		 (cond
		    ((terminated-thread-exception? (thread-%exc-result t))
		     (raise (thread-%exc-result t)))
		    (else
		     v))))))))

;*---------------------------------------------------------------------*/
;*    thread-await! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-await! s . arg)
   (let* ((t (current-thread)))
      ;; await until the signal is present
      (define (await e s)
	 (let ((env (scheduler-env s)))
	    (if (signal-lookup e env)
		;; the signal is present we return its value
		(signal-value e env)
		;; the signal is absent, we block the thread
		(let ((scdl (thread-scheduler t)))
		   (signal-register-thread! e env t)
		   (thread-set-await! t)
		   (if (and (%lazy-signal? e) (not (%lazy-signal-run e)))
		       (with-access::%lazy-signal e (run action scheduler id)
			  (set! run #t)
			  (thread-%detach! t id)
			  (action scheduler e)
			  (thread-%attach! t))
		       (thread-cooperate t))
		   (signal-value e env)))))
      ;; await at most n instants until the signal is present
      (define (await-ntimes e s timeout)
	 (if (and (number? timeout) (>= timeout 0))
	     (let ((env (scheduler-env s)))
		(if (signal-lookup e env)
		    (signal-value e env)
		    (begin
		       (signal-register-thread! e env t)
		       (thread-set-eoi! t (+fx 1 timeout))
		       (thread-set-await! t)
		       (thread-cooperate t)
		       (thread-unset-eoi! t)
		       (if (signal-lookup e env)
			   (signal-value e env)
			   #f))))
	     (error "thread-await" "Illegal timeout" timeout)))
      (cond
	 ((not (thread? t))
	  (error "thread-await" "no such thread" t))
	 ((not (thread-attached? t))
	  (error "thread-await" "unattached thread" t))
	 ((pair? arg)
	  (await-ntimes s (thread-scheduler t) (car arg)))
	 (else
	  (await s (thread-scheduler t))))))

;*---------------------------------------------------------------------*/
;*    thread-await*! ...                                               */
;*---------------------------------------------------------------------*/
(define (thread-await*! s* . arg)
   (let* ((t (current-thread)))
      ;; await until one of the signals is present
      (define (await* e* s)
	 (let ((env (scheduler-env s)))
	    (let loop ((es e*))
	       (cond
		  ((null? es)
		   (let ((scdl (thread-scheduler t)))
		      (for-each (lambda (e)
				   (signal-register-thread! e env t))
				e*)
		      (thread-set-await! t)
		      (thread-cooperate t)
		      (values (%thread-%awake-value t)
			      (%thread-%awake-signal t))))
		  ((signal-lookup (car es) env)
		   (values (signal-value (car es) env) (car es)))
		  (else
		   (loop (cdr es)))))))
      ;; await at most n instants until one of the signals is present
      (define (await*-ntimes e* s timeout)
	 (if (and (number? timeout) (>= timeout 0))
	     (let ((env (scheduler-env s)))
		(let loop ((es e*)
			   (stage 'init))
		   (cond
		      ((null? es)
		       (if (eq? stage 'init)
			   (let ((scdl (thread-scheduler t)))
			      (for-each (lambda (e)
					   (signal-register-thread! e env t))
					e*)
			      (thread-set-eoi! t (+fx 1 timeout))
			      (thread-set-await! t)
			      (thread-cooperate t)
			      (thread-unset-eoi! t)
			      (loop e* 'end))
			   (values #f #f)))
		      ((signal-lookup (car es) env)
		       (values (signal-value (car es) env) (car es)))
		      (else
		       (loop (cdr es) stage)))))
	     (error "thread-await*" "Illegal timeout" timeout)))
      (cond
	 ((not (thread? t))
	  (error "thread-await" "no such thread" t))
	 ((not (thread-attached? t))
	  (error "thread-await" "unattached thread" t))
	 ((pair? arg)
	  (await*-ntimes s* (thread-scheduler t) (car arg)))
	 (else
	  (await* s* (thread-scheduler t))))))

;*---------------------------------------------------------------------*/
;*    thread-yield/timeout! ...                                        */
;*---------------------------------------------------------------------*/
(define (thread-yield/timeout! timeout)
   (let ((t (current-thread)))
      (if (thread? t)
	  (with-access::thread t (scheduler)
	     ;; we now, really stop the thread
	     (thread-set-yield! t timeout)
	     (thread-cooperate t)))))

;*---------------------------------------------------------------------*/
;*    thread-sleep! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-sleep! timeout)
   (if (>fx timeout 0)
       (thread-yield/timeout! timeout)))

;*---------------------------------------------------------------------*/
;*    thread-yield! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-yield!)
   (thread-yield/timeout! 0))

;*---------------------------------------------------------------------*/
;*    thread-get-values! ...                                           */
;*---------------------------------------------------------------------*/
(define (thread-get-values! s)
   (let ((t (current-thread)))
      (if (thread? t)
	  (begin
	     (thread-yield!)
	     (signal-last-values s (scheduler-env (thread-scheduler t)))))))

;*---------------------------------------------------------------------*/
;*    thread-get-values*! ...                                          */
;*---------------------------------------------------------------------*/
(define (thread-get-values*! s*)
   (let ((t (current-thread)))
      (if (thread? t)
	  (begin
	     (thread-yield!)
	     (map (lambda (s)
		     (cons s (signal-last-values
			      s (scheduler-env (thread-scheduler t)))))
		  s*)))))

;*---------------------------------------------------------------------*/
;*    thread-await-values! ...                                         */
;*---------------------------------------------------------------------*/
(define (thread-await-values! s . tmt)
   (apply thread-await! s tmt)
   (thread-get-values! s))
   
;*---------------------------------------------------------------------*/
;*    thread-await-values*! ...                                        */
;*---------------------------------------------------------------------*/
(define (thread-await-values*! s* . tmt)
   (apply thread-await*! s* tmt)
   (thread-get-values*! s*))
   
;*---------------------------------------------------------------------*/
;*    object-display ::thread ...                                      */
;*---------------------------------------------------------------------*/
(define-method (object-display o::thread . port)
   (with-output-to-port (if (pair? port) (car port) (current-output-port))
      (lambda ()
	 (with-access::thread o (name)
	    (display* "#<thread:" name ">")))))

;*---------------------------------------------------------------------*/
;*    object-write ::thread ...                                        */
;*---------------------------------------------------------------------*/
(define-method (object-write o::thread . port)
   (with-output-to-port (if (pair? port) (car port) (current-output-port))
      (lambda ()
	 (with-access::thread o (name %state)
	    (display* "#<thread:" name ":" %state ">")))))

;*---------------------------------------------------------------------*/
;*    object-print ::thread ...                                        */
;*---------------------------------------------------------------------*/
(define-method (object-print o::thread port print-slot)
   (object-write o port))

;*---------------------------------------------------------------------*/
;*    current-thread ...                                               */
;*---------------------------------------------------------------------*/
(define (current-thread)
   (if (not (scheduler? (current-scheduler)))
       #f
       (%scheduler-current-thread (current-scheduler))))

;*---------------------------------------------------------------------*/
;*    thread-start! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-start! t . s)
   ;; check that the thread is not already attached to a scheduler
   ;; (i.e. the thread is not already running in another scheduler)
   (if (thread-attached? t)
       (error "thread-start!" "thread already running" t)
       (let ((scdl (get-scheduler s 'thread-start!)))
	  ;; attach the thread to the scheduler
	  (with-access::thread t (scheduler %state)
	     (set! scheduler scdl)
	     ;; start the builtin thread
	     (%bgl-thread-start (thread-%builtin t))
	     (set! %state 'runnable))
	  ;; adding a thread only appends it to the list of thread
	  ;; to be started at the next scheduler instant
	  (with-access::%scheduler scdl (tostart threads startn)
	     ;; increment the number of started threads
	     (set! startn (+fx 1 startn))
	     (set! threads (cons t threads))
	     ;; thread are append in the inverse order.
	     (set! tostart (cons t tostart)))
	  ;; return the started thread
	  t)))
   
;*---------------------------------------------------------------------*/
;*    thread-terminate! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-terminate! t)
   (if (thread-attached? t)
       (if (not (memq (thread-%state t) '(dying dead)))
	   (with-access::thread t (scheduler %state %exc-result)
	      ;; mark the termination
	      (set! %exc-result *terminated-thread-exception*)
	      ;; adding a thread simply append it to the list of thread
	      ;; to be stoped at the next scheduler instant
	      (with-access::%scheduler scheduler (toterminate)
		 ;; thread are append in the inverse order.
		 (set! toterminate (cons t toterminate)))))))

;*---------------------------------------------------------------------*/
;*    thread-suspend! ...                                              */
;*---------------------------------------------------------------------*/
(define (thread-suspend! t)
   (if (thread-attached? t)
       (with-access::thread t (%state)
	  (if (not (memq %state '(suspended dead)))
	      (let ((scdl (thread-scheduler t)))
		 ;; adding a thread simply append it to the list of thread
		 ;; to be stoped at the next scheduler instant
		 (with-access::%scheduler scdl (tosuspend)
		    ;; thread are append in the inverse order.
		    (if (not (memq t tosuspend))
			(set! tosuspend (cons t tosuspend)))))))))

;*---------------------------------------------------------------------*/
;*    thread-resume! ...                                               */
;*---------------------------------------------------------------------*/
(define (thread-resume! t)
   (if (thread-attached? t)
       (with-access::thread t (%state)
	  (if (not (eq? %state 'dead))
	      (let ((scdl (thread-scheduler t)))
		 ;; adding a thread simply append it to the list of thread
		 ;; to be stoped at the next scheduler instant
		 (with-access::%scheduler scdl (toresume)
		    ;; thread are append in the inverse order.
		    (if (not (memq t toresume))
			(set! toresume (cons t toresume)))))))))
