/*=====================================================================*/
/*    .../prgm/project/bigloo/fthread/src/Posix/bglasyncthunk.c        */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Feb 25 18:13:15 2002                          */
/*    Last change :  Sun May 18 06:49:36 2003 (serrano)                */
/*    Copyright   :  2002-03 Manuel Serrano                            */
/*    -------------------------------------------------------------    */
/*    The implementation of asynchronous thunk (non blocking IO).      */
/*=====================================================================*/
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <gc.h>

#define GC_PRIVATE_H
#include <bglthread.h>

/*---------------------------------------------------------------------*/
/*    Global lock                                                      */
/*---------------------------------------------------------------------*/
pthread_mutex_t async_lock;
pthread_cond_t async_cv;
static obj_t async_token = 0L;
extern char *strerror();

extern pthread_key_t bglkey;

/*---------------------------------------------------------------------*/
/*    bglthread_t                                                      */
/*    bglthread_async_new ...                                          */
/*---------------------------------------------------------------------*/
bglthread_t
bglthread_async_new( obj_t proc, obj_t scheduler, obj_t event, obj_t id ) {
   bglthread_t th = bglthread_new_with_name( proc, id );

   th->user1 = scheduler;
   th->user2 = event;
   return th;
}

/*---------------------------------------------------------------------*/
/*    static void *                                                    */
/*    bglthread_async_run ...                                          */
/*---------------------------------------------------------------------*/
static void *
bglthread_async_run( void *arg ) {
   bglthread_t th = (bglthread_t)arg;

   PROCEDURE_ENTRY( th->thunk )( th->thunk, th->user1, th->user2, BEOA );

   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_async_start ...                                        */
/*---------------------------------------------------------------------*/
void
bglthread_async_start( bglthread_t thread ) {
   pthread_attr_t a;

   pthread_attr_init( &a );
   pthread_attr_setdetachstate( &a, PTHREAD_CREATE_DETACHED );

   if( pthread_create( &(thread->pthread), &a, bglthread_async_run, thread ) )
      FAILURE( string_to_bstring( "thread-start!" ),
	       string_to_bstring( "Cannot start thread" ),
	       string_to_bstring( strerror( errno ) ) );
}

/*---------------------------------------------------------------------*/
/*    !!! UTILISER UN LOCK PAR SCHEDULER                               */
/*    (BGLTHREAD->CV ET BGLTHREAD->LOCK). !!!                          */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bglthread_async_scheduler_funcall ...                            */
/*---------------------------------------------------------------------*/
obj_t
bglthread_async_scheduler_funcall( obj_t thunk, obj_t token ) {
   obj_t res;
   
   pthread_mutex_lock( &async_lock );
   
   async_token = 0L;
   res = PROCEDURE_ENTRY( thunk )( thunk, token, BEOA );
   async_token = token;
   
   pthread_cond_signal( &async_cv );
   pthread_mutex_unlock( &async_lock );
   
   return res;
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bglthread_async_scheduler_funcall2 ...                           */
/*---------------------------------------------------------------------*/
obj_t
bglthread_async_scheduler_funcall2( obj_t thunk, obj_t token, obj_t val ) {
   obj_t res;
   
   pthread_mutex_lock( &async_lock );
   
   async_token = 0L;
   res = PROCEDURE_ENTRY( thunk )( thunk, token, val, BEOA );
   async_token = token;
   
   pthread_cond_signal( &async_cv );
   pthread_mutex_unlock( &async_lock );
   
   return res;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_async_await ...                                        */
/*---------------------------------------------------------------------*/
void
bglthread_async_await( obj_t t ) {
   while( async_token != t ) {
      pthread_cond_wait( &async_cv, &async_lock );
   }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bglthread_async_print ...                                        */
/*---------------------------------------------------------------------*/
obj_t
bglthread_async_print( obj_t o, obj_t port ) {
   if( OUTPUT_STRING_PORTP( port ) ) {
      display_string( o, port );
      strport_flush( port );
   } else {
      static int sig = 0;
      FILE *fout = OUTPUT_PORT( port ).file;
      long  len  = STRING_LENGTH( o );
      char *aux  = &STRING_REF( o, 0 );
      
      fwrite( aux, 1, len, fout );
      fflush( fout );
   }
   
   return o;
}
